# Emmy version 1.0.0
# EmmyParser.py
# by Antoine Rauzy
# Copyrights (c) 2019 AltaRica Association
#
# This file is part of Emmy, a pedagogical stochastic simulator.
# Emmy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Emmy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Emmy.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Models
#   3.1) Blocks
#   3.2) Elements

# 1) Imported modules
# -------------------

from S2MLCore import *

from EmmyModels import *
from EmmyLexer import *

# 2) Creator
# ----------

class EmmyParser(S2MLParser):
  def __init__(self, engine):
    S2MLParser.__init__(self, engine)

# 3) Models
# ---------

# 3.1) Blocks
# -----------

  def _ParseBlockElement(self, container):
    element = None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      return element
    elif token.GetType()==EmmyType.KEYWORD_ASSERTION:
      element = self._ParseAssertion(container)
    elif token.GetType()==EmmyType.KEYWORD_PARAMETER:
      element = self._ParseParameter(container)
    elif token.GetType()==EmmyType.KEYWORD_PROBABILITY:
      element = self._ParseProbabilityDeclaration(container)
    elif token.GetType()==EmmyType.KEYWORD_OBSERVER:
      element = self._ParseObserver(container)
    elif token.GetType()==EmmyType.KEYWORD_BLOCK:
      element = self._ParseBlock(container)
    elif token.GetType()==EmmyType.KEYWORD_EXTENDS:
      element = self._ParseExtendsDirective(container)
    elif token.GetType()==EmmyType.KEYWORD_EMBEDS:
      element = self._ParseEmbedsDirective(container)
    elif token.GetType()==EmmyType.KEYWORD_CLONES:
      element = self._ParseClonesDirective(container)
    elif token.GetType()==EmmyType.KEYWORD_BOOLEAN:
      element = self._ParseVariableDeclaration(container)
    elif token.GetType()==EmmyType.KEYWORD_INTEGER:
      element = self._ParseVariableDeclaration(container)
    elif token.GetType()==EmmyType.KEYWORD_REAL:
      element = self._ParseVariableDeclaration(container)
    elif token.GetType()==EmmyType.KEYWORD_SYMBOL:
      element = self._ParseVariableDeclaration(container)
    elif token.GetType()==EmmyType.TOKEN_LBRACE:
      element = self._ParseVariableDeclaration(container)
    elif token.GetType()==EmmyType.TOKEN_LBRACK:
      element = self._ParseVariableDeclaration(container)
    elif token.GetType()==EmmyType.TOKEN_LPAR:
      element = self._ParseFunctionCall(container)
    else:
      instantiatedElementIdentifier = self._ParseIdentifier()
      if instantiatedElementIdentifier==None:
        return None
      instantiatedElement = self._model.LookForIdentifiedElement(instantiatedElementIdentifier)
      if instantiatedElement==None:
        self._RaiseParseError(token, "Undeclared domain or class %s" % instantiatedElementIdentifier)
        return None
      self._currentTokenIndex -= 1
      if EmmyModel.IsDomainDeclaration(instantiatedElement):
        element = self._ParseVariableDeclaration(container)
      else:
        element = self._ParseClassInstance(container)
    return element

# 3.2) Elements
# -------------

  def _ParseDomainDeclaration(self, container):
    token = self._ParseGivenToken(EmmyType.KEYWORD_DOMAIN)
    if not self._model.CanComposeDomain(container):
      self._RaiseParseError(token, "Domains are not authorized here")
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      return None
    domain = self._ParseActualDomain()
    if domain==None:
      return None
    declaration = self._model.NewDomainDeclaration(identifier, domain, container)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.GetType()!=EmmyType.TOKEN_IDENTIFIER:
        break
      constant1 = self._ParseConstant()
      if constant1==None:
        return None
      token = self._ParseGivenToken(EmmyType.TOKEN_LT)
      if token==None:
        return None
      constant2 = self._ParseConstant()
      if constant2==None:
        return None
      declaration.AppendConstantOrder(constant1, constant2)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.GetType()!=EmmyType.TOKEN_COMMA:
        break
      self._currentTokenIndex += 1
    token = self._ParseGivenToken(EmmyType.TOKEN_SEMICOLON)
    if token==None:
      return None
    return declaration

  def _ParseVariableDeclaration(self, container):
    variables = []
    domain = self._ParseDomain()
    if domain==None:
      return None
    while True:
      path = self._ParsePath()
      if path==None:
        return None
      variable = self._model.NewVariable(domain, path, container)
      variables.append(variable)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==EmmyType.TOKEN_COMMA:
        self._currentTokenIndex += 1
      else:
        break
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==EmmyType.TOKEN_LPAR:
      variables = self._ParseAttributes(variables)
      if variables==None:
        return None
    token = self._ParseGivenToken(EmmyType.TOKEN_SEMICOLON)
    if token==None:
      return None
    return variables

  def _ParseAssertion(self, container):
    token = self._ParseGivenToken(EmmyType.KEYWORD_ASSERTION)
    if token==None:
      return None
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or (token.GetType()!=EmmyType.TOKEN_IDENTIFIER and token.GetType()!=EmmyType.TOKEN_PATH):
        break
      assignment = self._ParseAssignment(container)
      if assignment==None:
        self._RecoverError(EmmyType.KEYWORD_END)
        return None
    return container

  def _ParseAssignment(self, container):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==EmmyType.TOKEN_LPAR:
      return self._ParseConnection(container)
    variablePath = self._ParsePath()
    token = self._ParseGivenToken(EmmyType.TOKEN_ASSIGN)
    if token==None:
      return None
    expression = self._ParseStochasticExpression()
    if expression==None:
      return None
    token = self._ParseGivenToken(EmmyType.TOKEN_SEMICOLON)
    if token==None:
      return None
    return self._model.NewAssignment(variablePath, expression, container)

  def _ParseProbabilityDeclaration(self, container):
    if not self._model.CanComposeProbabilityDeclaration(container):
      self._RaiseParseError(token, "Probability declarations are not authorized here")
      return None
    token = self._ParseGivenToken(EmmyType.KEYWORD_PROBABILITY)
    if token==None:
      return None
    variablePath = self._ParsePath()
    if variablePath==None:
      return None
    constant = self._ParseConstant()
    if constant==None:
      return None
    token = self._ParseGivenToken(EmmyType.TOKEN_EQUAL)
    if token==None:
      return None
    distribution = self._ParseStochasticExpression()
    if distribution==None:
      return None
    token = self._ParseGivenToken(EmmyType.TOKEN_SEMICOLON)
    if token==None:
      return None
    declaration = self._model.NewProbabilityDeclaration(variablePath, constant, distribution, container)
    return declaration

  def _ParseObserver(self, container):
    if not self._model.CanComposeObserver(container):
      self._RaiseParseError(token, "Observer declarations are not authorized here")
      return None
    token = self._ParseGivenToken(EmmyType.KEYWORD_OBSERVER)
    if token==None:
      return None
    path = self._ParsePath()
    if path==None:
      return None
    token = self._ParseGivenToken(EmmyType.TOKEN_EQUAL)
    if token==None:
      return None
    definition = self._ParseStochasticExpression()
    if definition==None:
      return None
    token = self._ParseGivenToken(EmmyType.TOKEN_SEMICOLON)
    if token==None:
      return None
    observer = self._model.NewObserver(path, definition, container)
    return observer
