# S2MLCore version 1.0.0
# S2MLConstant.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Constants
#  2.1) Creators
#  2.2) Assignments and conversions
#  2.3) Boolean operations
#  2.4) Arithmetic operations
#  2.5) Inequalities
#  2.6) String operations

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLReal import *

# 2) Constants
# ------------

# 2.1) Creators
# -------------

class S2MLConstant:
  NONE = 0
  BOOLEAN = 1
  INTEGER = 2
  REAL = 3
  SYMBOL = 4
  STRING = 5
  ANY = 6

  def __init__(self, type, value):
    self._type = type
    self._value = value

  @staticmethod
  def NewNone():
    return S2MLConstant(S2MLConstant.NONE, None)

  @staticmethod
  def NewBoolean(value):
    return S2MLConstant(S2MLConstant.BOOLEAN, value)

  @staticmethod
  def NewInteger(value):
    return S2MLConstant(S2MLConstant.INTEGER, value)

  @staticmethod
  def NewReal(value):
    return S2MLConstant(S2MLConstant.REAL, S2MLReal.New(value))

  @staticmethod
  def NewPosInf():
    return S2MLConstant(S2MLConstant.REAL, S2MLReal.PosInf())
    
  @staticmethod
  def NewNegInf():
    return S2MLConstant(S2MLConstant.REAL, S2MLReal.NegInf())

  @staticmethod
  def NewSymbol(value):
    return S2MLConstant(S2MLConstant.SYMBOL, value)

  @staticmethod
  def NewString(value):
    return S2MLConstant(S2MLConstant.STRING, value)

  @staticmethod
  def Copy(other):
    return S2MLConstant(other._type, other._value)

# 2.2) Assignments and conversions
# --------------------------------

  def GetType(self):
    return self._type
  def GetValue(self):
    return self._value

  def IsNone(self):
    return self._type==S2MLConstant.NONE
  def IsBoolean(self):
    return self._type==S2MLConstant.BOOLEAN
  def IsFalse(self):
    return self._type==S2MLConstant.BOOLEAN and not self._value
  def IsTrue(self):
    return self._type==S2MLConstant.BOOLEAN and self._value
  def IsInteger(self):
    return self._type==S2MLConstant.INTEGER
  def IsReal(self):
    return self._type==S2MLConstant.REAL
  def IsNumber(self):
    return self.IsInteger() or self.IsReal()
  def IsSymbol(self):
    return self._type==S2MLConstant.SYMBOL
  def IsString(self):
    return self._type==S2MLConstant.STRING

  def Set(self, other):
    self._type = other._type
    self._value = other._value
  def Reset(self):
    self._type = S2MLConstant.NONE
    self._value = None
  def SetAsBoolean(self, value):
    self._type = S2MLConstant.BOOLEAN
    self._value = value
  def SetAsInteger(self, value):
    self._type = S2MLConstant.INTEGER
    self._value = value
  def SetAsReal(self, value):
    self._type = S2MLConstant.REAL
    self._value = S2MLReal.New(value)
  def SetAsPosInf(self):
    self._type = S2MLConstant.REAL
    self._value = S2MLReal.PosInf()
  def SetAsNegInf(self):
    self._type = S2MLConstant.REAL
    self._value = S2MLReal.NegInf()
  def SetAsSymbol(self, value):
    self._type = S2MLConstant.SYMBOL
    self._value = value
  def SetAsString(self, value):
    self._type = S2MLConstant.STRING
    self._value = value

  def ToString(self):
    if self._type==S2MLConstant.NONE:
      return "None"
    if self._type==S2MLConstant.BOOLEAN:
      if self._value:
        return "true"
      return "false"
    if self._type==S2MLConstant.INTEGER:
      return str(self._value)
    if self._type==S2MLConstant.REAL:
      return S2MLReal.ToString(self._value)
    if self._type==S2MLConstant.SYMBOL:
      return self._value
    if self._type==S2MLConstant.STRING:
      return self._value
    return None

  @staticmethod
  def FromString(self, string):
    if string==None:
      return S2MLConstant.NewNone()
    return S2MLConstant.NewString(string)

  def ToFloat(self):
    if self.IsInteger():
      return S2MLReal.New(self._value)
    if self.IsReal():
      return self._value
    return S2MLReal.New(0)

# 2.3) Boolean operations
# -----------------------

  def AndMe(self, other):
    if self.IsBoolean():
      if other.IsBoolean():
        self._value = self._value and other._value
      else:
        self.Reset()
    else:
      self.Reset()

  def OrMe(self, other):
    if self.IsBoolean():
      if other.IsBoolean():
        self._value = self._value or other._value
      else:
        self.Reset()
    else:
      self.Reset()

  def NotMe(self, other):
    if self.IsBoolean():
      if other.IsBoolean():
        self._value = not other._value
      else:
        self.Reset()
    else:
      self.Reset()

  def CountMe(self, other):
    if self.IsNumber():
      if other.IsBoolean():
        if other._value:
          self._value += 1
      else:
        self.Reset()
    else:
      self.Reset()

# 2.4) Arithmetic operations
# --------------------------

  def AddMe(self, other):
    if self.IsInteger():
      if other.IsInteger():
        self._value += other._value
      elif other.IsReal():
        self._type = S2MLConstant.REAL
        self._value = S2MLReal.Add(S2MLReal.New(self._value), other._value)
      else:
        self.Reset()
    elif self.IsReal():
      if other.IsInteger():
        self._value = S2MLReal.Add(self._value, S2MLReal.New(other._value))
      elif other.IsReal():
        self._value = S2MLReal.Add(self._value, other._value)
      else:
        self.Reset()
    else:
      self.Reset()

  def SubMe(self, other):
    if self.IsInteger():
      if other.IsInteger():
        self._value -= other._value
      elif other.IsReal():
        self._type = S2MLConstant.REAL
        self._value = S2MLReal.Sub(S2MLReal.New(self._value), other._value)
      else:
        self.Reset()
    elif self.IsReal():
      if other.IsInteger():
        self._value = S2MLReal.Sub(self._value, S2MLReal.New(other._value))
      elif other.IsReal():
        self._value = S2MLReal.Sub(self._value, other._value)
      else:
        self.Reset()
    else:
      self.Reset()

  def MulMe(self, other):
    if self.IsInteger():
      if other.IsInteger():
        self._value *= other._value
      elif other.IsReal():
        self._type = S2MLConstant.REAL
        self._value = S2MLReal.Mul(S2MLReal.New(self._value), other._value)
      else:
        self.Reset()
    elif self.IsReal():
      if other.IsInteger():
        self._value = S2MLReal.Mul(self._value, S2MLReal.New(other._value))
      elif other.IsReal():
        self._value = S2MLReal.Mul(self._value, other._value)
      else:
        self.Reset()
    else:
      self.Reset()

  def DivMe(self, other):
    if self.IsInteger():
      if other.IsInteger():
        self._value /= other._value
      elif other.IsReal():
        self._type = S2MLConstant.REAL
        self._value = S2MLReal.Div(S2MLReal.New(self._value), other._value)
      else:
        self.Reset()
    elif self.IsReal():
      if other.IsInteger():
        self._value = S2MLReal.Div(self._value, S2MLReal.New(other._value))
      elif other.IsReal():
        self._value = S2MLReal.Div(self._value, other._value)
      else:
        self.Reset()
    else:
      self.Reset()

  def NegMe(self, other):
    if other.IsInteger():
      self._type = S2MLConstant.INTEGER
      self._value = -other._value
    elif other.IsReal():
      self._type = S2MLConstant.REAL
      self._value = S2MLReal.Neg(other._value)
    else:
      self.Reset()

  @staticmethod
  def Add(arg1, arg2):
    result = S2MLConstant.Copy(arg1)
    result.AddMe(arg2)
    return result

  @staticmethod
  def Sub(arg1, arg2):
    result = S2MLConstant.Copy(arg1)
    result.SubMe(arg2)
    return result

  @staticmethod
  def Mul(arg1, arg2):
    result = S2MLConstant.Copy(arg1)
    result.MulMe(arg2)
    return result

  @staticmethod
  def Div(arg1, arg2):
    result = S2MLConstant.Copy(arg1)
    result.DivMe(arg2)
    return result

  @staticmethod
  def Neg(arg1):
    result = S2MLConstant.Copy(arg1)
    result.NegMe()
    return result

  def MinMe(self, other):
    if self.IsInteger():
      if other.IsInteger():
        if other._value<self._value:
          self._value = other._value
      elif other.IsReal():
        if S2MLReal.Lt(other._value, S2MLReal.New(self._value)):
          self._type = S2MLConstant.REAL
          self._value = other._value
      else:
        self.Reset()
    elif self.IsReal():
      if other.IsInteger():
        if S2MLReal.Lt(S2MLReal.New(other._value), self._value):
          self._type = S2MLConstant.INTEGER
          self._value = other._value
      elif other.IsReal():
        if S2MLReal.Lt(other._value, self._value):
          self._value = other._value
      else:
        self.Reset()
    else:
      self.Reset()

  def MaxMe(self, other):
    if self.IsInteger():
      if other.IsInteger():
        if other._value>self._value:
          self._value = other._value
      elif other.IsReal():
        if S2MLReal.Gt(other._value, S2MLReal.New(self._value)):
          self._type = S2MLConstant.REAL
          self._value = other._value
      else:
        self.Reset()
    elif self.IsReal():
      if other.IsInteger():
        if S2MLReal.Gt(S2MLReal.New(other._value), self._value):
          self._type = S2MLConstant.INTEGER
          self._value = other._value
      elif other.IsReal():
        if S2MLReal.Gt(other._value, self._value):
          self._value = other._value
      else:
        self.Reset()
    else:
      self.Reset()

# 2.5) Inequalities
# -----------------

  @staticmethod
  def AreComparable(arg1, arg2):
    if arg1.IsNone() and arg2.IsNone():
      return True
    if arg1.IsBoolean() and arg2.IsBoolean():
      return True
    if arg1.IsNumber() and arg2.IsNumber():
      return True
    if arg1.IsSymbol() and arg2.IsSymbol():
      return True
    if arg1.IsString() and arg2.IsString():
      return True
    return False

  @staticmethod
  def IsEq(arg1, arg2):
    if arg1.IsNone() and arg2.IsNone():
      return True
    if arg1.IsBoolean() and arg2.IsBoolean():
      return arg1._value==arg2._value
    if arg1.IsInteger():
      if arg2.IsInteger():
        return arg1._value==arg2._value
      if arg2.IsReal():
        return S2MLReal.Eq(S2MLReal.New(arg1._value), arg2._value)
      return False
    if arg1.IsReal():
      if arg2.IsInteger():
        return S2MLReal.Eq(arg1._value, S2MLReal.New(arg2._value))
      if arg2.IsReal():
        return S2MLReal.Eq(arg1._value, arg2._value)
      return False
    if arg1.IsSymbol() and arg2.IsSymbol():
      return arg1._value==arg2._value
    if arg1.IsString() and arg2.IsString():
      return arg1._value==arg2._value
    return False

  @staticmethod
  def IsDf(arg1, arg2):
    if arg1.IsNone() and arg2.IsNone():
      return False
    if arg1.IsBoolean() and arg2.IsBoolean():
      return arg1._value!=arg2._value
    if arg1.IsInteger():
      if arg2.IsInteger():
        return arg1._value!=arg2._value
      if arg2.IsReal():
        return S2MLReal.Df(S2MLReal.New(arg1._value), arg2._value)
      return True
    if arg1.IsReal():
      if arg2.IsInteger():
        return S2MLReal.Df(arg1._value, S2MLReal.New(arg2._value))
      if arg2.IsReal():
        return S2MLReal.Df(arg1._value, arg2._value)
      return True
    if arg1.IsSymbol() and arg2.IsSymbol():
      return arg1._value!=arg2._value
    if arg1.IsString() and arg2.IsString():
      return arg1._value!=arg2._value
    return True

  @staticmethod
  def IsLt(arg1, arg2):
    if arg1.IsNone() and arg2.IsNone():
      return False
    if arg1.IsBoolean() and arg2.IsBoolean():
      return False
    if arg1.IsInteger():
      if arg2.IsInteger():
        return arg1._value<arg2._value
      if arg2.IsReal():
        return S2MLReal.Lt(S2MLReal.New(arg1._value), arg2._value)
      return False
    if arg1.IsReal():
      if arg2.IsInteger():
        return S2MLReal.Lt(arg1._value, S2MLReal.New(arg2._value))
      if arg2.IsReal():
        return S2MLReal.Lt(arg1._value, arg2._value)
      return False
    if arg1.IsSymbol() and arg2.IsSymbol():
      return arg1._value<arg2._value
    if arg1.IsString() and arg2.IsString():
      return arg1._value<arg2._value
    return False

  @staticmethod
  def IsGt(arg1, arg2):
    if arg1.IsNone() and arg2.IsNone():
      return False
    if arg1.IsBoolean() and arg2.IsBoolean():
      return False
    if arg1.IsInteger():
      if arg2.IsInteger():
        return arg1._value>arg2._value
      if arg2.IsReal():
        return S2MLReal.Gt(S2MLReal.New(arg1._value), arg2._value)
      return False
    if arg1.IsReal():
      if arg2.IsInteger():
        return S2MLReal.Gt(arg1._value, S2MLReal.New(arg2._value))
      if arg2.IsReal():
        return S2MLReal.Gt(arg1._value, arg2._value)
      return False
    if arg1.IsSymbol() and arg2.IsSymbol():
      return arg1._value>arg2._value
    if arg1.IsString() and arg2.IsString():
      return arg1._value>arg2._value
    return False

  @staticmethod
  def IsLeq(arg1, arg2):
    if arg1.IsNone() and arg2.IsNone():
      return False
    if arg1.IsBoolean() and arg2.IsBoolean():
      return False
    if arg1.IsInteger():
      if arg2.IsInteger():
        return arg1._value<=arg2._value
      if arg2.IsReal():
        return S2MLReal.Leq(S2MLReal.New(arg1._value), arg2._value)
      return False
    if arg1.IsReal():
      if arg2.IsInteger():
        return S2MLReal.Leq(arg1._value, S2MLReal.New(arg2._value))
      if arg2.IsReal():
        return S2MLReal.Leq(arg1._value, arg2._value)
      return False
    if arg1.IsSymbol() and arg2.IsSymbol():
      return arg1._value<=arg2._value
    if arg1.IsString() and arg2.IsString():
      return arg1._value<=arg2._value
    return False

  @staticmethod
  def IsGeq(arg1, arg2):
    if arg1.IsNone() and arg2.IsNone():
      return False
    if arg1.IsBoolean() and arg2.IsBoolean():
      return False
    if arg1.IsInteger():
      if arg2.IsInteger():
        return arg1._value>=arg2._value
      if arg2.IsReal():
        return S2MLReal.Geq(S2MLReal.New(arg1._value), arg2._value)
      return False
    if arg1.IsReal():
      if arg2.IsInteger():
        return S2MLReal.Geq(arg1._value, S2MLReal.New(arg2._value))
      if arg2.IsReal():
        return S2MLReal.Geq(arg1._value, arg2._value)
      return False
    if arg1.IsSymbol() and arg2.IsSymbol():
      return arg1._value>=arg2._value
    if arg1.IsString() and arg2.IsString():
      return arg1._value>=arg2._value
    return False

  @staticmethod
  def Eq(arg1, arg2):
    return S2MLConstant.NewBoolean(S2MLConstant.IsEq(arg1, arg2))

  @staticmethod
  def Df(arg1, arg2):
    return S2MLConstant.NewBoolean(S2MLConstant.IsDf(arg1, arg2))

  @staticmethod
  def Lt(arg1, arg2):
    return S2MLConstant.NewBoolean(S2MLConstant.IsLt(arg1, arg2))

  @staticmethod
  def Gt(arg1, arg2):
    return S2MLConstant.NewBoolean(S2MLConstant.IsGt(arg1, arg2))

  @staticmethod
  def Leq(arg1, arg2):
    return S2MLConstant.NewBoolean(S2MLConstant.IsLeq(arg1, arg2))

  @staticmethod
  def Geq(arg1, arg2):
    return S2MLConstant.NewBoolean(S2MLConstant.IsGeq(arg1, arg2))

  def EqMe(self, other1, other2):
    if self.IsBoolean():
      self._value = S2MLConstant.IsEq(other1, other2)
    else:
      self.Reset()

  def DfMe(self, other1, other2):
    if self.IsBoolean():
      self._value = S2MLConstant.IsDf(other1, other2)
    else:
      self.Reset()

  def LtMe(self, other1, other2):
    if self.IsBoolean():
      self._value = S2MLConstant.IsLt(other1, other2)
    else:
      self.Reset()

  def GtMe(self, other1, other2):
    if self.IsBoolean():
      self._value = S2MLConstant.IsGt(other1, other2)
    else:
      self.Reset()

  def LeqMe(self, other1, other2):
    if self.IsBoolean():
      self._value = S2MLConstant.IsLeq(other1, other2)
    else:
      self.Reset()

  def GeqMe(self, other1, other2):
    if self.IsBoolean():
      self._value = S2MLConstant.IsGeq(other1, other2)
    else:
      self.Reset()

# 2.6) String operations
# ----------------------

