# S2MLCore version 1.0.0
# S2MLInterpreter.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Environment variables
# 3) Interpreter
# 4) Scripts
# 5) Command reset
# 6) Command load
# 7) Command instantiate
# 8) Command print
# 9) Command compute
# 10) Command set
# 11) Command profile
# 12) Methods to parse the command line

# 1) Imported modules
# -------------------

from S2MLMatrices import *
from S2MLStatistics import *

# 2) Creator
# ----------

class S2MLInterpreter:
  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._printer = engine.GetPrinter()
    self._inputFileName = None
    self._inputFile = sys.stdin
    self._errors = 0
    self._line = None
    self._lineNumber = 0
    self._environmentVariables = dict()
    self.InitializeEnvironmentVariables()
    self._arguments = []
    self._argumentIndex = 0
    self._listLevel = 0
    self._commandOutputFileName = None
    self._commandOutputMode = "w"

  def _RaiseError(self, message):
    self._errors += 1
    self._engine.RaiseError(message)

# 3) Environment variables
# ------------------------

  def GetEnvironmentVariables(self):
    return self._environmentVariables
  def GetIdentifiersOfEnvironmentVariables(self):
    return self._environmentVariables.keys()
  def GetValueOfEnvironmentVariable(self, identifier):
    return self._environmentVariables.get(identifier, None)
  def SetValueOfEnvironmentVariable(self, identifier, constant):
    self._environmentVariables[identifier] = constant

  def InitializeEnvironmentVariables(self):
    pass

# 4) Scripts
# ----------

  def LoadScript(self, fileName):
    try:
      file = open(fileName, "r")
    except:
      self._RaiseError('Unable to open file "%s"\n' % fileName)
      return 1
    self.ParseFile(fileName, file)
    file.close()
    return self._errors

  def ParseFile(self, fileName, file):
    previousInputFileName = self._inputFileName
    previousFile = self._inputFile
    previousLine = self._line
    previousLineNumber = self._lineNumber
    self._inputFileName = fileName
    self._inputFile = file
    self._line = None
    self._lineNumber = 0
    self._ParseCommands()
    self._inputFileName = previousInputFileName
    self._inputFile = previousFile
    self._line = previousLine
    self._lineNumber = previousLineNumber

  def ParseCommand(self, commandText):
    previousFile = self._inputFile
    previousLine = self._line
    previousLineNumber = self._lineNumber
    self._inputFile = None
    self._line = commandText
    self._lineNumber = 0
    self._ParseCommandLine()
    if self._arguments!=[]:
      self._ParseCommand()
    self._inputFile = previousFile
    self._line = previousLine
    self._lineNumber = previousLineNumber

  def _ParseCommands(self):
    from S2MLEngine import S2MLEngine
    for self._line in self._inputFile:
      if self._engine.GetExecutionStatus()!=S2MLEngine.ACTIVE:
        break
      self._ParseCommandLine()
      if self._arguments!=[]:
        self._ParseCommand()

  def _ParseCommand(self):
    self._commandOutputFileName = None
    self._commandOutputMode = "w"
    commandName = self.GetArgument()
    if commandName=="":
      print("uh?")
    elif commandName=="instantiate":
      self._errors += self._ParseCommandInstantiate()
    elif commandName=="load":
      self._errors += self._ParseCommandLoad()
    elif commandName=="print":
      self._errors += self._ParseCommandPrint()
    elif commandName=="set":
      self._errors += self._ParseCommandSet()
    elif commandName=="profile":
      self._errors += self._ParseCommandProfile()
    else:
      self._RaiseError("line %d\nUnknow command %s\n" % (self._lineNumber, commandName))

# 5) Command reset
# ---------------

  # reset model
  def _ParseCommandReset(self):
    what = self.GetArgument()
    if what!="model":
      self._RaiseError("line %d\nIncorrect argument %s\n" % (self._lineNumber, what))
      return 1
    fileName = self.GetArgument()
    if fileName==None:
      self._RaiseError("line %d\nA file name was expected\n" % self._lineNumber)
      return 1
    if what=="model":
      self._engine.ResetModel()
      return self._errors
    return 1 # impossible

# 6) Command load
# ---------------

  # load model fileName
  # load script fileName
  def _ParseCommandLoad(self):
    what = self.GetArgument()
    if what!="model" and what!="script":
      self._RaiseError("line %d\nIncorrect argument %s\n" % (self._lineNumber, what))
      return 1
    fileName = self.GetArgument()
    if fileName==None:
      self._RaiseError("line %d\nA file name was expected\n" % self._lineNumber)
      return 1
    if what=="model":
      self._engine.ImportModel(fileName)
      return self._errors
    if what=="script":
      self._engine.LoadScript(fileName)
      return self._errors
    return 1 # impossible

# 7) Command instantiate
# ----------------------

  # instantiate model
  def _ParseCommandInstantiate(self):
    what = self.GetArgument()
    if what!="model":
      self._RaiseError("line %d\nIncorrect argument %s\n" % (self._lineNumber, what))
      return 1
    if what=="model":
      return self._engine.InstantiateModel()
    return 1 # impossible

# 8) Command print
# ----------------

  # print model [output=fileName mode=(append|write)]
  # print instantiated-model [output=fileName mode=(append|write)]
  # print variables [output=fileName mode=(append|write)]
  # print variable variableName [output=fileName mode=(append|write)]
  # print profiles [output=fileName mode=(append|write)]
  # print profile profileName [output=fileName mode=(append|write)]
  def _ParseCommandPrint(self):
    what = self.GetArgument()
    self.GetCommandOutputFileOptions()
    if what=="model":
      if self._commandOutputFileName==None:
        error = self._engine.PrintModel()
      else:
        error = self._engine.ExportModel(self._commandOutputFileName, mode=self._commandOutputMode)
      return error
    if what=="instantiated-model":
      if self._commandOutputFileName==None:
        error = self._engine.PrintInstantiatedModel()
      else:
        error = self._engine.ExportInstantiatedModel(self._commandOutputFileName, mode=self._commandOutputMode)
      return error
    if what=="variables":
      if self._commandOutputFileName==None:
        error = self._engine.PrintEnvironmentVariables()
      else:
        error = self._engine.ExportEnvironmentVariables(self._commandOutputFileName, mode=self._commandOutputMode)
      return error
    if what=="variable":
      variableName = self.GetArgument()
      if variableName==None:
        self._RaiseError("line %d\nA variable name was expected\n" % self._lineNumber)
        return 1
      if self._commandOutputFileName==None:
        error = self._engine.PrintEnvironmentVariable(variableName)
      else:
        error = self._engine.ExportEnvironmentVariable(variableName, self._commandOutputFileName, mode=self._commandOutputMode)
      return error
    if what=="profiles":
      if self._commandOutputFileName==None:
        error = self._engine.PrintIndicatorProfiles()
      else:
        error = self._engine.ExportIndicatorProfiles(self._commandOutputFileName, mode=self._commandOutputMode)
      return error
    if what=="profile":
      profileName = self.GetArgument()
      if profileName==None:
        self._RaiseError("line %d\nA profile name was expected\n" % self._lineNumber)
        return 1
      if self._commandOutputFileName==None:
        error = self._engine.PrintIndicatorProfile(profileName)
      else:
        error = self._engine.ExportIndicatorProfile(profileName, self._commandOutputFileName, mode=self._commandOutputMode)
      return error
    self._RaiseError("line %d\nIncorrect argument %s\n" % (self._lineNumber, what))
    return 1

# 9) Command compute
# ------------------

  def _ParseCommandCompute(self):
    what = self.GetArgument()
    self._RaiseError("line %d\nSyntax error\n" % self._lineNumber)
    return 1

# 10) Command set
# ---------------

  # set block blockName
  def _ParseCommandSet(self):
    what = self.GetArgument()
    if what==None:
      self._RaiseError("line %d\nMissing argument to command set\n" % self._lineNumber)
      return 1
    if what=="block":
      return self._ParseCommandSetBlock(self)
    if not self.IsIdentifier(what):
      self._RaiseError("line %d\nA variable identifier was expected (%s)\n" % (self._lineNumber, what))
      return 1
    identifier = what
    value = self.GetArgument()
    if value==None:
      self._RaiseError("line %d\nA variable value was expected\n" % self._lineNumber)
      return 1
    if value=="false":
      constant = S2MLConstant.NewBoolean(False)
    elif value=="true":
      constant = S2MLConstant.NewBoolean(True)
    elif self.IsInt(value):
      constant = S2MLConstant.NewInteger(int(value))
    elif self.IsFloat(value):
      constant = S2MLConstant.NewReal(float(value))
    elif self.IsIdentifier(value):
      constant = S2MLConstant.NewSymbol(value)
    elif self.IsString(value):
      constant = S2MLConstant.NewString(value)
    else:
      self._RaiseError("line %d\nIncorrect variable value\n" % (self._lineNumber, argument))
      return 1
    self.SetValueOfEnvironmentVariable(identifier, constant)
    return 1 # impossible

  # set block blockName
  def _ParseCommandSetBlock(self):
    what = self.GetArgument()
    if what!="block":
      self._RaiseError("line %d\nIncorrect argument %s\n" % (self._lineNumber, what))
      return 1
    blockName = self.GetArgument()
    if blockName==None:
      self._RaiseError("line %d\nA block name was expected\n" % self._lineNumber)
      return 1
    if what=="block":
      return self._engine.SetTargetBlock(blockName)
    return 1 # impossible

# 11) Command profile
# -------------------

  def _ParseCommandProfile(self):
    what = self.GetArgument()
    if what=="new":
      sourceName = self.GetArgument()
      if sourceName==None:
        self._RaiseError("line %d\nA profile name was expected\n" % self._lineNumber)
        return 1
      return self._engine.NewIndicatorProfile(sourceName)
    if what=="clone":
      sourceName = self.GetArgument()
      if sourceName==None:
        self._RaiseError("line %d\nA profile name was expected\n" % self._lineNumber)
        return 1
      targetName = self.GetArgument()
      if targetName==None:
        self._RaiseError("line %d\nA profile name was expected\n" % self._lineNumber)
        return 1
      return self._engine.CloneIndicatorProfile(sourceName, targeName)
    if what=="delete":
      sourceName = self.GetArgument()
      if sourceName==None:
        self._RaiseError("line %d\nA profile name was expected\n" % self._lineNumber)
        return 1
      return self._engine.DeleteIndicatorProfile(sourceName)
    if what=="set":
      argument = self.GetArgument()
      if argument==None:
        self._RaiseError("line %d\nAn option of the command 'profile set' was expected\n" % self._lineNumber)
        return 1
      sourceName = self.GetArgument()
      if sourceName==None:
        self._RaiseError("line %d\nA profile name was expected\n" % self._lineNumber)
        return 1
      if argument=="mean":
        boolValue = self.GetBool()
        if boolValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionMean(sourceName, boolValue)
      if argument=="standard-deviation":
        boolValue = self.GetBool()
        if boolValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionStandardDeviation(sourceName, boolValue)
      if argument=="confidence-interval-90":
        boolValue = self.GetBool()
        if boolValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionConfidenceInterval90(sourceName, boolValue)
      if argument=="confidence-interval-95":
        boolValue = self.GetBool()
        if boolValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionConfidenceInterval95(sourceName, boolValue)
      if argument=="confidence-interval-99":
        boolValue = self.GetBool()
        if boolValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionConfidenceInterval99(sourceName, boolValue)
      if argument=="extrema":
        boolValue = self.GetBool()
        if boolValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionExtrema(sourceName, boolValue)
      if argument=="quantiles":
        boolValue = self.GetBool()
        if boolValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionQuantiles(sourceName, boolValue)
      if argument=="number-of-quantiles":
        intValue = self.GetInt()
        if intValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionNumberOfQuantiles(sourceName, intValue)
      if argument=="distribution":
        boolValue = self.GetBool()
        if boolValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionDistribution(sourceName, boolValue)
      if argument=="cumulative-distribution":
        boolValue = self.GetBool()
        if boolValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionCumulativeDistribution(sourceName, boolValue)
      if argument=="number-of-points":
        intValue = self.GetInt()
        if intValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionNumberOfPoints(sourceName, intValue)
      if argument=="bins":
        boolValue = self.GetBool()
        if boolValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionBins(sourceName, boolValue)
      if argument=="number-of-bins":
        intValue = self.GetInt()
        if intValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionNumberOfBins(sourceName, intValue)
      if argument=="number-of-stored-values":
        intValue = self.GetInt()
        if intValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionNumberOfStoredValues(sourceName, intValue)
      if argument=="bootstrapping":
        boolValue = self.GetBool()
        if boolValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionBootstrapping(sourceName, boolValue)
      if argument=="number-of-samples":
        intValue = self.GetInt()
        if intValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionNumberOfSamples(sourceName, intValue)
      if argument=="sampling-ratio":
        floatValue = self.GetFloat()
        if floatValue==None:
          return 1
        return self._engine.SetIndicatorProfileOptionSamplingRatio(sourceName, floatValue)
      self._RaiseError("line %d\nIncorrect argument %s of the commant profile set\n" % (self._lineNumber, argument))
      return 1
    self._RaiseError("line %d\nIncorrect argument %s\n" % (self._lineNumber, what))
    return 1

# 12) Methods to parse the command line
# -------------------------------------

  def _ParseCommandLine(self):
    self._lineNumber += 1
    self._line.rstrip("\r\n")
    self._line.rstrip("\n")
    self._arguments = []
    state = 1
    column = 0
    while True:
      if state==1:
        if column>=len(self._line):
          if self._listLevel==0:
            break
          if self._inputFile==None:
            return
          self._line = self._inputFile.readline()
          self._lineNumber += 1
          self._line.rstrip("\r\n")
          self._line.rstrip("\n")
          column = 0
        else:
          c = self._line[column]
          if c.isspace():
            pass
          elif c=="#":
            break
          elif c=="=":
            self._arguments.append("=")
          elif c==",":
            self._arguments.append(",")
          elif c=="[":
            self._arguments.append("[")
            self._listLevel += 1
          elif c=="]":
            self._arguments.append("]")
            self._listLevel -= 1
          elif c=="'":
            start = column+1
            state = 3
          elif c=='"':
            start = column+1
            state = 4
          elif c=="\\":
            if self._inputFile==None:
              return
            self._line = self._inputFile.readline()
            self._lineNumber += 1
            self._line.rstrip("\r\n")
            self._line.rstrip("\n")
            column = 0
          else:
            start = column
            state = 2
          column += 1
      elif state==2:
        if column>=len(self._line):
          argument = self._line[start:column]
          self._arguments.append(argument)
          break
        else:
          c = self._line[column]
          if c.isspace() or c=="#" or c=="=" or c=="'" or c=='"' or c=="," or c=="[" or c=="]":
            argument = self._line[start:column]
            self._arguments.append(argument)
            state = 1
          else:
            column += 1
      elif state==3:
        if column>=len(self._line):
          self._RaiseError("Line %d, unterminated string" % self._lineNumber)
          self._arguments = []
          return
        else:
          c = self._line[column]
          if c=="'":
            argument = self._line[start:column]
            self._arguments.append(argument)
            state = 1
          column += 1
      elif state==4:
        if column>=len(self._line):
          self._RaiseError("Line %d, unterminated string" % self._lineNumber)
          self._arguments = []
          return
        else:
          c = self._line[column]
          if c=='"':
            argument = self._line[start:column]
            self._arguments.append(argument)
            state = 1
          column += 1
      self._argumentIndex = 0
    return

  def LookUpArgument(self):
    if self._argumentIndex>=len(self._arguments):
      return None
    return self._arguments[self._argumentIndex]

  def GetArgument(self):
    if self._argumentIndex>=len(self._arguments):
      return None
    argument = self._arguments[self._argumentIndex]
    self._argumentIndex += 1
    return argument

  def PutBackArgument(self):
    self._argumentIndex -= 1

  def GetStringOption(self, name, defaultValue):
    argument = self.LookUpArgument()
    if argument!=name:
      return defaultValue
    if self._argumentIndex+2>=len(self._arguments):
      self._RaiseError("Line %d. Wrong number of arguments" % self._lineNumber)
      return None
    if self._arguments[self._argumentIndex+1]!="=":
      self._RaiseError("Line %d. A '=' was expected" % self._lineNumber)
      return None
    option = self._arguments[self._argumentIndex+2]
    self._argumentIndex += 3
    return option

  def GetBoolOption(self, name, defaultValue):
    argument = self.LookUpArgument()
    if argument!=name:
      return defaultValue
    if self._argumentIndex+2>=len(self._arguments):
      self._RaiseError("Line %d. Wrong number of arguments" % self._lineNumber)
      return None
    if self._arguments[self._argumentIndex+1]!="=":
      self._RaiseError("Line %d. A '=' was expected" % self._lineNumber)
      return None
    option = self._arguments[self._argumentIndex+2].lower()
    if option!="true" and option!="false":
      self._RaiseError("Line %d. A Boolean (true or false) was expected" % self._lineNumber)
      return None
    self._argumentIndex += 3
    return option=="true"

  def GetIntOption(self, name, defaultValue):
    argument = self.LookUpArgument()
    if argument!=name:
      return defaultValue
    if self._argumentIndex+2>=len(self._arguments):
      self._RaiseError("Line %d. Wrong number of arguments" % self._lineNumber)
      return None
    if self._arguments[self._argumentIndex+1]!="=":
      self._RaiseError("Line %d. A '=' was expected" % self._lineNumber)
      return None
    string = self._arguments[self._argumentIndex+2]
    if not self.IsInt(string):
      self._RaiseError("Line %d. An integer was expected" % self._lineNumber)
      return None
    option = int(string)
    self._argumentIndex += 3
    return option

  def GetFloatOption(self, name, defaultValue):
    argument = self.LookUpArgument()
    if argument!=name:
      return defaultValue
    if self._argumentIndex+2>=len(self._arguments):
      self._RaiseError("Line %d. Wrong number of arguments" % self._lineNumber)
      return None
    if self._arguments[self._argumentIndex+1]!="=":
      self._RaiseError("Line %d. A '=' was expected" % self._lineNumber)
      return None
    string = self._arguments[self._argumentIndex+2]
    if not self.IsFloat(string):
      self._RaiseError("Line %d. A float was expected" % self._lineNumber)
      return None
    option = float(string)
    self._argumentIndex += 3
    return option

  def GetCommandOutputFileOptions(self):
    while True:
      argument = self.LookUpArgument()
      if argument=="output":
        self.GetOutputFileOption()
      elif argument=="mode":
        self.GetModeOption("w")
      else:
        break

  def GetOutputFileOption(self):
    argument = self.LookUpArgument()
    if argument!="output":
      return None
    if self._argumentIndex+2>=len(self._arguments):
      self._RaiseError("Line %d. Wrong number of arguments" % self._lineNumber)
      return None
    if self._arguments[self._argumentIndex+1]!="=":
      self._RaiseError("Line %d. A '=' was expected" % self._lineNumber)
      return None
    fileName = self._arguments[self._argumentIndex+2]
    if fileName==None:
      self._RaiseError("line %d\nA file name was expected\n" % self._lineNumber)
      return None
    self._argumentIndex += 3
    self._commandOutputFileName = fileName
    return fileName

  def GetMissionTimeOption(self, defaultValue):
    argument = self.LookUpArgument()
    if argument!="mission-time" and argument!="mission-times":
      return defaultValue
    if self._argumentIndex+2>=len(self._arguments):
      self._RaiseError("Line %d. Wrong number of arguments" % self._lineNumber)
      return None
    if self._arguments[self._argumentIndex+1]!="=":
      self._RaiseError("Line %d. A '=' was expected" % self._lineNumber)
      return None
    self._argumentIndex += 2
    missionTimes = self.GetFloatList()
    return missionTimes

  def GetModeOption(self, defaultValue):
    argument = self.LookUpArgument()
    if argument!="mode":
      return defaultValue
    if self._argumentIndex+2>=len(self._arguments):
      self._RaiseError("Line %d. Wrong number of arguments" % self._lineNumber)
      return None
    if self._arguments[self._argumentIndex+1]!="=":
      self._RaiseError("Line %d. A '=' was expected" % self._lineNumber)
      return None
    mode = self._arguments[self._argumentIndex+2]
    if mode=="write":
      mode = "w"
    if mode=="append":
      mode = "a"
    if mode!="w" and mode!="a":
      return None
    self._argumentIndex += 3
    self._commandOutputMode = mode
    return mode

  def GetIntList(self):
    items = []
    argument = self.LookUpArgument()
    if argument=="[":
      argument = self.GetArgument()
      items = []
      while True:
        item = self.GetInt()
        if item==None:
          return None
        items.append(item)
        argument = self.GetArgument()
        if argument=="]":
          break
        elif argument==",":
          pass
        else:
          self._RaiseError("Line %d. Syntax error" % self._lineNumber)
          return None
    else:
      item = self.GetInt()
      if item==None:
        return None
      items = [item]
    return items

  def GetFloatList(self):
    items = []
    argument = self.LookUpArgument()
    if argument=="[":
      argument = self.GetArgument()
      items = []
      while True:
        item = self.GetFloat()
        if item==None:
          return None
        items.append(item)
        argument = self.GetArgument()
        if argument=="]":
          break
        elif argument==",":
          pass
        else:
          self._RaiseError("Line %d. Syntax error" % self._lineNumber)
          return None
    else:
      item = self.GetFloat()
      if item==None:
        return None
      items = [item]
    return items

  def GetBool(self):
    argument = self.GetArgument()
    if argument==None:
      self._RaiseError("Line %d. A Boolean was expected" % self._lineNumber)
      return None
    if argument!="true" and argument!="false":
      self._RaiseError("Line %d. A Boolean (true or false) was expected" % self._lineNumber)
      return None
    return argument=="true"

  def GetInt(self):
    argument = self.GetArgument()
    if argument==None:
      self._RaiseError("Line %d. An int was expected" % self._lineNumber)
      return None
    if not self.IsInt(argument):
      self._RaiseError("Line %d. An int was expected" % self._lineNumber)
      return None
    return int(argument)

  def GetFloat(self):
    argument = self.GetArgument()
    if argument==None:
      self._RaiseError("Line %d. A float was expected" % self._lineNumber)
      return None
    if not self.IsFloat(argument):
      self._RaiseError("Line %d. A float was expected" % self._lineNumber)
      return None
    return float(argument)

  def IsInt(self, item):
    try:
      value = int(item)
      return True
    except ValueError:
      return False

  def IsFloat(self, item):
    try:
      value = float(item)
      return True
    except ValueError:
      return False

  def IsIdentifier(self, item):
    return re.match(r'[a-zA-Z_][a-zA-Z0-9_-]*', item)

  def IsString(self, item):
    return re.match(r'".*"', item) or re.match(r"'.*'", item)
